#include <opencl/dsa.h>
#include <opencl/rsa.h>
#include <opencl/nr.h>
#include <opencl/dh.h>
#include <opencl/rw.h>

#include <opencl/look_pk.h>
#include <opencl/rng.h>
using namespace OpenCL;

#include <iostream>
#include <fstream>
#include <ctime>
#include <string>

#define DEBUG 0

BigInt to_bigint(const std::string&);

RSA_PrivateKey load_rsa_key(const std::string&);
RW_PrivateKey  load_rw_key(const std::string&);

void bench_enc(PK_Encryptor*, const std::string&, double, bool);
void bench_dec(PK_Encryptor*, PK_Decryptor*, const std::string&, double, bool);
void bench_sig(PK_Signer*, const std::string&, double, bool);
void bench_ver(PK_Signer*, PK_Verifier*, const std::string&, double, bool);
void bench_kas(PK_Key_Agreement_Scheme*, const std::string&, double, bool);

void bench_rsa_set(RSA_PrivateKey& key, const std::string keybits,
                   double seconds, bool html)
   {
   bench_enc(get_pk_encryptor(key, "Raw"),
             "RSA-" + keybits, seconds, html);
   bench_dec(get_pk_encryptor(key, "Raw"),
             get_pk_decryptor(key, "Raw"),
             "RSA-" + keybits, seconds, html);
   }

void bench_rw_set(RW_PrivateKey& key, const std::string keybits,
                  double seconds, bool html)
   {
   bench_ver(get_pk_signer(key, "EMSA2(SHA-1)"),
             get_pk_verifier(key, "EMSA2(SHA-1)"),
             "RW-" + keybits, seconds, html);
   bench_sig(get_pk_signer(key, "EMSA2(SHA-1)"),
             "RW-" + keybits, seconds, html);
   }

void bench_dsa_set(DSA_PrivateKey& key, const std::string keybits,
                   double seconds, bool html)
   {
   bench_ver(get_pk_signer(key, "EMSA1(SHA-1)"),
             get_pk_verifier(key, "EMSA1(SHA-1)"),
             "DSA-" + keybits, seconds, html);
   bench_sig(get_pk_signer(key, "EMSA1(SHA-1)"),
             "DSA-" + keybits, seconds, html);
   }

void bench_nr_set(NR_PrivateKey& key, const std::string keybits,
                  double seconds, bool html)
   {
   bench_ver(get_pk_signer(key, "EMSA1(SHA-1)"),
             get_pk_verifier(key, "EMSA1(SHA-1)"),
             "NR-" + keybits, seconds, html);
   bench_sig(get_pk_signer(key, "EMSA1(SHA-1)"),
             "NR-" + keybits, seconds, html);
   }

void bench_dh(DH_PrivateKey& key, const std::string keybits,
              double seconds, bool html)
   {
   bench_kas(get_pk_kas(key, "Raw"),
             "DH-" + keybits, seconds, html);
   }

void bench_pk(bool html, double seconds)
   {
   try {

   DH_PrivateKey dh768(get_dl_group("IETF-768"));
   DH_PrivateKey dh1024(get_dl_group("IETF-1024"));
   DH_PrivateKey dh1536(get_dl_group("IETF-1536"));
   DH_PrivateKey dh2048(get_dl_group("IETF-2048"));
   bench_dh(dh768,  "768",  seconds, html);
   bench_dh(dh1024, "1024", seconds, html);
   bench_dh(dh1536, "1536", seconds, html);
   bench_dh(dh2048, "2048", seconds, html);

   DSA_PrivateKey dsa512(get_dl_group("DSA-512"));
   DSA_PrivateKey dsa768(get_dl_group("DSA-768"));
   DSA_PrivateKey dsa1024(get_dl_group("DSA-1024"));
   bench_dsa_set(dsa512,  "512",  seconds, html);
   bench_dsa_set(dsa768,  "768",  seconds, html);
   bench_dsa_set(dsa1024, "1024", seconds, html);

   NR_PrivateKey nr512(get_dl_group("DSA-512"));
   NR_PrivateKey nr768(get_dl_group("DSA-768"));
   NR_PrivateKey nr1024(get_dl_group("DSA-1024"));
   bench_nr_set(nr512,  "512",  seconds, html);
   bench_nr_set(nr768,  "768",  seconds, html);
   bench_nr_set(nr1024, "1024", seconds, html);

   RSA_PrivateKey rsa512 = load_rsa_key("checks/keys/rsa512.key");
   RSA_PrivateKey rsa1024 = load_rsa_key("checks/keys/rsa1024.key");
   RSA_PrivateKey rsa2048 = load_rsa_key("checks/keys/rsa2048.key");
   bench_rsa_set(rsa512,  "512",  seconds, html);
   bench_rsa_set(rsa1024, "1024", seconds, html);
   bench_rsa_set(rsa2048, "2048", seconds, html);

   RW_PrivateKey rw512 = load_rw_key("checks/keys/rw512.key");
   RW_PrivateKey rw1024 = load_rw_key("checks/keys/rw1024.key");
   bench_rw_set(rw512,  "512",  seconds, html);
   bench_rw_set(rw1024, "1024", seconds, html);

   }
   catch(OpenCL::Exception& e)
      {
      std::cout << "Exception caught: " << e.what() << std::endl;
      return;
      }
   catch(std::exception& e)
      {
      std::cout << "Standard library exception caught: "
                << e.what() << std::endl;
      return;
      }
   catch(...)
      {
      std::cout << "Unknown exception caught." << std::endl;
      return;
      }

   }

void print_result(bool, u32bit runs, u32bit clocks_used,
                  const std::string& algo_name, const std::string& op)
   {
   std::cout << algo_name << ": ";

   double seconds = (double)clocks_used / CLOCKS_PER_SEC;
   double mseconds_per_run = 1000 * (seconds / runs);

   std::cout.setf(std::ios::fixed, std::ios::floatfield);
   std::cout.precision(2);
   std::cout << mseconds_per_run << " ms / " << op << std::endl;
   }

void bench_enc(PK_Encryptor* enc, const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(msg, MSG_SIZE);
      enc->encrypt(msg, MSG_SIZE);
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "public operation");
   }

void bench_dec(PK_Encryptor* enc, PK_Decryptor* dec,
               const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];
   Global_RNG::randomize(msg, MSG_SIZE);

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   SecureVector<byte> encrypted_msg = enc->encrypt(msg, MSG_SIZE);

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      dec->decrypt(encrypted_msg, encrypted_msg.size());
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "private operation");
   }

void bench_sig(PK_Signer* sig, const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(msg, MSG_SIZE);
      sig->update(msg, MSG_SIZE);
      sig->signature();
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "private operation");
   }

void bench_ver(PK_Signer* sig, PK_Verifier* ver,
               const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];
   Global_RNG::randomize(msg, MSG_SIZE);

   sig->update(msg, MSG_SIZE);
   SecureVector<byte> signature = sig->signature();

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      ver->update(msg, MSG_SIZE);
      ver->valid_signature(signature, signature.size());
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "public operation");
   }

void bench_kas(PK_Key_Agreement_Scheme* kas, const std::string& algo_name,
               double seconds, bool html)
   {
   /* 128 bits: should always be considered valid (what about ECC?) */
   static const u32bit REMOTE_KEY_SIZE = 16;
   byte key[REMOTE_KEY_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(key, REMOTE_KEY_SIZE);
      kas->derive_key(key, REMOTE_KEY_SIZE, 0);
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "key agreement");
   }

/*************************************************
* Key loading procedures                         *
*************************************************/
RSA_PrivateKey load_rsa_key(const std::string& file)
   {
   std::ifstream keyfile(file.c_str());
   if(!keyfile)
      throw Exception("Couldn't open the RSA key file " + file);

   std::string e, p, q;

   std::getline(keyfile, e);
   std::getline(keyfile, p);
   std::getline(keyfile, q);

   RSA_PrivateKey key(to_bigint(p), to_bigint(q), to_bigint(e));

#if DEBUG
   if(!key.check_params())
      throw Exception("Bad key from " + file);
#endif
   return key;
   }

RW_PrivateKey load_rw_key(const std::string& file)
   {
   std::ifstream keyfile(file.c_str());
   if(!keyfile)
      throw Exception("Couldn't open the RW key file " + file);

   std::string e, p, q;

   std::getline(keyfile, e);
   std::getline(keyfile, p);
   std::getline(keyfile, q);

   RW_PrivateKey key(to_bigint(p), to_bigint(q), to_bigint(e));

#if DEBUG
   if(!key.check_params())
      throw Exception("Bad key from " + file);
#endif
   return key;
   }
